@@ -6,7 +6,7 @@ JSON::Validator - Validate data against a JSON schema
 
 =head1 VERSION
 
-0.52
+0.53
 
 =head1 SYNOPSIS
 
@@ -93,7 +93,7 @@ use constant DEBUG => $ENV{JSON_VALIDATOR_DEBUG} || $ENV{SWAGGER2_DEBUG} || 0;
 use constant WARN_ON_MISSING_FORMAT => $ENV{JSON_VALIDATOR_WARN_ON_MISSING_FORMAT}
   || $ENV{SWAGGER2_WARN_ON_MISSING_FORMAT} ? 1 : 0;
 
-our $VERSION = '0.52';
+our $VERSION = '0.53';
 our @EXPORT_OK = qw( validate_json );
 
 sub E { bless {path => $_[0] || '/', message => $_[1]}, 'JSON::Validator::Error'; }
@@ -454,42 +454,42 @@ sub _resolve_schema {
 
 sub _validate {
   my ($self, $data, $path, $schema) = @_;
-  my $check_all = grep { $schema->{$_} } qw( allOf oneOf );
-  my (@errors, @types);
+  my $type = $schema->{type} || _guess_schema_type($schema, $data);
+  my @errors;
 
-  push @types, map { $schema->{$_} } grep { $schema->{$_} } qw( allOf anyOf oneOf );
-  push @types, $schema->{type} || _guess_schema_type($schema) || 'any' unless @types;
+  # Test base schema before allOf, anyOf or oneOf
+  if (ref $type eq 'ARRAY') {
+    for my $type (@$type) {
+      my $method = sprintf '_validate_type_%s', $type;
+      my @e = $self->$method($data, $path, $schema);
+      warn "[JSON::Validator] type @{[$path||'/']} => $method [@e]\n" if DEBUG == 2;
+      push @errors, \@e;
+      next if @e;
+      @errors = ();
+      last;
+    }
+  }
+  elsif ($type) {
+    my $method = sprintf '_validate_type_%s', $type;
+    @errors = $self->$method($data, $path, $schema);
+    warn "[JSON::Validator] type @{[$path||'/']} $method [@errors]\n" if DEBUG == 2;
+    return @errors if @errors;
+  }
 
-  #warn Carp::longmess(Data::Dumper::Dumper([$data, $path, $schema]));
-  #$SIG{__WARN__} = sub { Carp::confess(Data::Dumper::Dumper($schema)) };
-
-  if ($schema->{not}) {
-    @errors = $self->_validate($data, $path, $schema->{not});
+  if (my $rules = $schema->{not}) {
+    push @errors, $self->_validate($data, $path, $rules);
+    warn "[JSON::Validator] not @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
     return @errors ? () : (E $path, 'Should not match.');
   }
 
-  for my $t (map { ref $_ eq 'ARRAY' ? @$_ : $_ } @types) {
-    $t //= 'null';
-    if (ref $t eq 'HASH') {
-      push @errors, [$self->_validate($data, $path, $t)];
-      return if !$check_all and !@{$errors[-1]};    # valid
-    }
-    elsif (my $code = $self->can(sprintf '_validate_type_%s', $t)) {
-      push @errors, [$self->$code($data, $path, $schema)];
-      return if !$check_all and !@{$errors[-1]};    # valid
-    }
-    elsif ($t eq 'file') {
-      return;                                       # Skip validating raw file
-    }
-    else {
-      return E $path, "Cannot validate type '$t'";
-    }
+  if (my $rules = $schema->{allOf}) {
+    push @errors, $self->_validate_all_of($data, $path, $rules);
   }
-
-  if ($schema->{oneOf}) {
-    my $n = grep { @$_ == 0 } @errors;
-    return if $n == 1;                              # one match
-    return E $path, "Expected only one to match." if $n == @errors;
+  elsif ($rules = $schema->{anyOf}) {
+    push @errors, $self->_validate_any_of($data, $path, $rules);
+  }
+  elsif ($rules = $schema->{oneOf}) {
+    push @errors, $self->_validate_one_of($data, $path, $rules);
   }
 
   if (@errors > 1) {
@@ -517,28 +517,62 @@ sub _validate {
     }
   }
 
-  return @{$errors[0]};
+  return @{$errors[0]} if @errors and ref $errors[0] eq 'ARRAY';
+  return @errors;
 }
 
-sub _validate_additional_properties {
-  my ($self, $data, $path, $schema) = @_;
-  my $properties = $schema->{additionalProperties};
+sub _validate_all_of {
+  my ($self, $data, $path, $rules) = @_;
   my @errors;
 
-  if (ref $properties eq 'HASH') {
-    push @errors, $self->_validate($_, $path, $properties) for values %$data;
-  }
-  elsif (!$properties) {
-    my @keys = grep { $_ !~ /^(description|id|title)$/ } keys %$data;
-    if (@keys) {
-      local $" = ', ';
-      push @errors, E $path, "Properties not allowed: @keys.";
-    }
+  for my $rule (@$rules) {
+    push @errors, [$self->_validate($data, $path, $rule)];
   }
 
+  warn "[JSON::Validator] allOf @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
   return @errors;
 }
 
+sub _validate_any_of {
+  my ($self, $data, $path, $rules) = @_;
+  my @errors;
+
+  for my $rule (@$rules) {
+    my @e = $self->_validate($data, $path, $rule);
+    last unless @e;
+    push @errors, \@e;
+  }
+
+  if (@errors < @$rules) {
+    warn "[JSON::Validator] anyOf @{[$path||'/']} == success\n" if DEBUG == 2;
+    return;
+  }
+  else {
+    warn "[JSON::Validator] anyOf @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
+    return @errors;
+  }
+}
+
+sub _validate_one_of {
+  my ($self, $data, $path, $rules) = @_;
+  my $failed = 0;
+  my @errors;
+
+  for my $rule (@$rules) {
+    my @e = $self->_validate($data, $path, $rule);
+    $failed++ if @e;
+    push @errors, @e;
+  }
+
+  if ($failed + 1 == @$rules) {
+    warn "[JSON::Validator] oneOf @{[$path||'/']} == success\n" if DEBUG == 2;
+    return;
+  }
+
+  warn "[JSON::Validator] oneOf @{[$path||'/']} == failed=$failed/@{[int @$rules]} / @errors\n" if DEBUG == 2;
+  return E $path, 'Expected only one to match.';
+}
+
 sub _validate_type_enum {
   my ($self, $data, $path, $schema) = @_;
   my $enum = $schema->{enum};
@@ -552,6 +586,9 @@ sub _validate_type_enum {
   return E $path, "Not in enum list: @$enum.";
 }
 
+# TODO: Need to figure out if this is a Swagger specific thing
+sub _validate_type_file { }
+
 sub _validate_format {
   my ($self, $value, $path, $schema) = @_;
   my $code = $self->formats->{$schema->{format}};
@@ -565,53 +602,6 @@ sub _validate_format {
   return E $path, "Does not match $schema->{format} format.";
 }
 
-sub _validate_pattern_properties {
-  my ($self, $data, $path, $schema) = @_;
-  my $properties = $schema->{patternProperties};
-  my @errors;
-
-  for my $pattern (keys %$properties) {
-    my $v = $properties->{$pattern};
-    for my $tk (keys %$data) {
-      next unless $tk =~ /$pattern/;
-      push @errors, $self->_validate(delete $data->{$tk}, _path($path, $tk), $v);
-    }
-  }
-
-  return @errors;
-}
-
-sub _validate_properties {
-  my ($self, $data, $path, $schema) = @_;
-  my $properties = $schema->{properties};
-  my $required   = $schema->{required};
-  my (@errors, %required);
-
-  if ($required and ref $required eq 'ARRAY') {
-    $required{$_} = 1 for @$required;
-  }
-
-  for my $name (keys %$properties) {
-    my $p = $properties->{$name};
-    if (exists $data->{$name}) {
-      my $v = delete $data->{$name};
-      push @errors, $self->_validate_type_enum($v, $path, $p) if $p->{enum};
-      push @errors, $self->_validate($v, _path($path, $name), $p);
-    }
-    elsif ($p->{default}) {
-      $data->{$name} = $p->{default};
-    }
-    elsif ($required{$name}) {
-      push @errors, E _path($path, $name), "Missing property.";
-    }
-    elsif (_is_true($p->{required}) eq '1') {
-      push @errors, E _path($path, $name), "Missing property.";
-    }
-  }
-
-  return @errors;
-}
-
 sub _validate_type_any {
   return;
 }
@@ -663,8 +653,7 @@ sub _validate_type_array {
   elsif (ref $schema->{items} eq 'HASH') {
     for my $i (0 .. @$data - 1) {
       if ($schema->{items}{properties}) {
-        my $input = ref $data->[$i] eq 'HASH' ? {%{$data->[$i]}} : $data->[$i];
-        push @errors, $self->_validate_properties($input, "$path/$i", $schema->{items});
+        push @errors, $self->_validate_type_object($data->[$i], "$path/$i", $schema->{items});
       }
       else {
         push @errors, $self->_validate($data->[$i], "$path/$i", $schema->{items});
@@ -732,29 +721,57 @@ sub _validate_type_number {
 
 sub _validate_type_object {
   my ($self, $data, $path, $schema) = @_;
-  my @errors;
+  my ($additional, @errors, %rules);
 
   if (ref $data ne 'HASH') {
     return E $path, _expected(object => $data);
   }
-
-  # make sure _validate_xxx() does not mess up original $data
-  $data = {%$data};
-
   if (defined $schema->{maxProperties} and $schema->{maxProperties} < keys %$data) {
     push @errors, E $path, sprintf 'Too many properties: %s/%s.', int(keys %$data), $schema->{maxProperties};
   }
   if (defined $schema->{minProperties} and $schema->{minProperties} > keys %$data) {
     push @errors, E $path, sprintf 'Not enough properties: %s/%s.', int(keys %$data), $schema->{minProperties};
   }
-  if ($schema->{properties}) {
-    push @errors, $self->_validate_properties($data, $path, $schema);
+
+  while (my ($k, $r) = each %{$schema->{properties}}) {
+    push @{$rules{$k}}, $r if exists $data->{$k} or ($r->{required} and ref $r->{required} ne 'ARRAY');
   }
-  if ($schema->{patternProperties}) {
-    push @errors, $self->_validate_pattern_properties($data, $path, $schema);
+  while (my ($p, $r) = each %{$schema->{patternProperties}}) {
+    push @{$rules{$_}}, $r for grep { $_ =~ /$p/ } keys %$data;
   }
-  if (exists $schema->{additionalProperties}) {
-    push @errors, $self->_validate_additional_properties($data, $path, $schema);
+
+  # special case used internally
+  $rules{id} ||= [{type => 'string'}] if !$path and $data->{id};
+  $additional = exists $schema->{additionalProperties} ? $schema->{additionalProperties} : {};
+
+  if ($additional) {
+    $additional = {} unless ref $additional eq 'HASH';
+    $rules{$_} ||= [$additional] for keys %$data;
+  }
+  elsif (my @keys = grep { !$rules{$_} } keys %$data) {
+    local $" = ', ';
+    return E $path, "Properties not allowed: @keys.";
+  }
+
+  if (ref $schema->{required} eq 'ARRAY') {
+    for my $k (@{$schema->{required}}) {
+      push @{$rules{$k}}, {required => 1};
+    }
+  }
+
+  for my $k (keys %rules) {
+    for my $r (@{$rules{$k}}) {
+      if (!exists $data->{$k} and exists $schema->{default}) {
+        $data->{$k} = $r->{default};
+      }
+      if ($r->{required} and !exists $data->{$k}) {
+        push @errors, E _path($path, $k), 'Missing property.';
+      }
+      else {
+        push @errors, $self->_validate_type_enum($data->{$k}, _path($path, $k), $r) if $r->{enum};
+        push @errors, $self->_validate($data->{$k}, _path($path, $k), $r);
+      }
+    }
   }
 
   return @errors;
@@ -817,26 +834,29 @@ sub _guess_data_type {
   return lc $ref if $ref and !$blessed;
   return 'null' if !defined;
   return 'boolean' if $blessed and "$_" eq "1" or "$_" eq "0";
-  return 'integer' if /^\d+$/;
   return 'number' if B::svref_2object(\$_)->FLAGS & (B::SVp_IOK | B::SVp_NOK) and 0 + $_ eq $_ and $_ * 0 == 0;
   return $blessed || 'string';
 }
 
 sub _guess_schema_type {
-  return 'object' if $_[0]->{additionalProperties};
-  return 'object' if $_[0]->{patternProperties};
-  return 'object' if $_[0]->{properties};
-  return 'object' if defined $_[0]->{maxProperties} or defined $_[0]->{minProperties};
-  return 'array'  if $_[0]->{additionalItems};
-  return 'array'  if $_[0]->{items};
-  return 'array'  if $_[0]->{uniqueItems};
-  return 'array'  if defined $_[0]->{maxItems} or defined $_[0]->{minItems};
-  return 'string' if $_[0]->{pattern};
-  return 'string' if defined $_[0]->{maxLength} or defined $_[0]->{minLength};
-  return 'number' if $_[0]->{multipleOf};
-  return 'number' if defined $_[0]->{maximum} or defined $_[0]->{minimum};
-  return 'enum'   if $_[0]->{enum};
-  return;
+  return _guessed_right($_[1], 'object') if $_[0]->{additionalProperties};
+  return _guessed_right($_[1], 'object') if $_[0]->{patternProperties};
+  return _guessed_right($_[1], 'object') if $_[0]->{properties};
+  return _guessed_right($_[1], 'object') if defined $_[0]->{maxProperties} or defined $_[0]->{minProperties};
+  return _guessed_right($_[1], 'array')  if $_[0]->{additionalItems};
+  return _guessed_right($_[1], 'array')  if $_[0]->{items};
+  return _guessed_right($_[1], 'array')  if $_[0]->{uniqueItems};
+  return _guessed_right($_[1], 'array')  if defined $_[0]->{maxItems} or defined $_[0]->{minItems};
+  return _guessed_right($_[1], 'string') if $_[0]->{pattern};
+  return _guessed_right($_[1], 'string') if defined $_[0]->{maxLength} or defined $_[0]->{minLength};
+  return _guessed_right($_[1], 'number') if $_[0]->{multipleOf};
+  return _guessed_right($_[1], 'number') if defined $_[0]->{maximum} or defined $_[0]->{minimum};
+  return 'enum' if $_[0]->{enum};
+  return undef;
+}
+
+sub _guessed_right {
+  return _guess_data_type($_[0]) eq $_[1] ? $_[1] : undef;
 }
 
 sub _is_byte_string { $_[0] =~ /^[A-Za-z0-9\+\/\=]+$/; }
